<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <title>immuneML: Failed Analysis</title>
    <style>
        {{{css_style}}}
    </style>
    <meta name="viewport" content="width=device-width, initial-scale=1">

</head>
<body>
    <div class="container">
        <h1>Failed immuneML analysis</h1>
        <p>immuneML version: {{immuneML_version}}</p>

        <p>The immuneML run failed with an error. Please check the exception and log file below.
        You can also download the Archive for this immuneML run from the Galaxy History to inspect all files
        and intermediate results that were generated by immuneML.</p>

        <details>
        <summary>View exception</summary>
        <pre>{{exception}}</pre>
        </details>

        <details>
        <summary>View log file content</summary>
        <pre>{{logfile_content}}</pre>
        </details>
</body>
</html>