<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <title>immuneML: Subsampling Overview</title>
    <style>
        {{{css_style}}}
    </style>
    <meta name="viewport" content="width=device-width, initial-scale=1">

</head>
<body>
    <div class="container">
        <div>
            <h1>Subsampling {{name}}</h1>
            <p>immuneML version: {{immuneML_version}}</p>
            <p>Full specification is available <a href="{{{full_specs}}}">here</a>.</p>
            <p>Log file is available <a href="{{{logfile}}}">here</a>.</p>
            <p></p>
            <div class="card">
                <div class="col-container">
                    <div class="col">
                        <h3>Dataset information (original full dataset)</h3>
                        <h4>General information</h4>
                        <div class="table-container">
                            <table>
                            <tr>
                                <td><b>Name</b></td>
                                <td>{{full_dataset_name}}</td>
                            </tr>
                            <tr>
                                <td><b>Type</b></td>
                                <td>{{full_dataset_type}}</td>
                            </tr>
                            <tr>
                                <td><b>Locus</b></td>
                                <td>{{full_dataset_locus}}</td>
                            </tr>
                            <tr>
                                <td><b>Dataset size</b></td>
                                <td>{{full_dataset_size}}</td>
                            </tr>
                        </table>
                        </div>
                        <h4>Dataset labels</h4>
                        {{#show_full_dataset_labels}}
                        <p>The following label(s) are available for this dataset, and may for example be used for classification (note: the same labels are available for the subsampled datasets):</p>

                        <div class="table-container">
                            <table>
                                <thead>
                                <tr>
                                    <th><b>Label name</b></th>
                                    <th><b>Label classes</b></th>
                                </tr>
                                </thead>
                                <tbody>
                                {{#full_dataset_labels}}
                                <tr>
                                    <td>{{full_dataset_label_name}}</td>
                                    <td>{{full_dataset_label_classes}}</td>
                                </tr>
                                {{/full_dataset_labels}}
                                </tbody>
                            </table>
                        </div>
                        {{/show_full_dataset_labels}}
                        {{^show_full_dataset_labels}}
                        <p>This dataset does not have any available labels.</p>
                        {{/show_full_dataset_labels}}
                    </div>
                    <div class="col">
                        <h3>Subsampling results</h3>
                        {{#subsampled_datasets}}
                        <div>
                            <h5>Subsampled dataset {{sub_dataset_iter}}:</h5>
                            <div class="padded-left">Dataset name: {{sub_dataset_name}}</div>
                            <div class="padded-left">Dataset size: {{sub_dataset_size}}</div>
                            <div class="padded-left">
                                <div><span>Download formats: </span>
                                    <div>
                                        <ul>
                                        {{#formats}}
                                            <li><a href="{{dataset_download_link}}" download>{{format_name}}</a></li>
                                        {{/formats}}
                                        </ul>
                                    </div>
                                </div>
                            </div>
                        </div>
                        {{/subsampled_datasets}}
                    </div>
                </div>
            </div>
        </div>
    </div>
</body>
</html>